# coding: utf-8
"""
**pronto**: a Python frontend to ontologies
===========================================

"""



__version__='0.1.10'
__author__='Martin Larralde'
__author_email__ = 'martin.larralde@ens-cachan.fr'

from pronto.ontology import Ontology
from pronto.term import Term, TermList
from pronto.relationship import Relationship
from pronto.parser import Parser

__all__ = ["Ontology", "Term", "TermList", "Relationship", "Parser"]



# coding: utf-8

RSHIPS = ('has_regexp', 'has_order', 'has_units', 'has_domain', 'is_a', 'part_of', 'is_part')
RSHIP_INVERSE = {'is_a': 'can_be', 'is_part':'has_part', 'part_of': 'has_part'}


class Relationship(object):
    """
    A Relationship object.

    The Relationship class actually behaves as a factory, creating new
    relationships via the default Python syntax only if no relationship
    of the same name are present in the class py:attribute:: _instances
    (a dictionnary containing memoized relationships).

    """

    _instances = {}

    def __init__(self, obo_name, symmetry=None, transitivity=None,
                 reflexivity=None, complementary=None, prefix=None,
                 direction=None, comment=None, aliases=None):
        """Instantiate a new relationship.

        Parameters:
            obo_name (str): the name of the relationship as it appears
                in obo files (such as is_a, has_part, etc.)
            symetry (bool or None): the symetry of the relationship
            transitivity (bool or None): the transitivity of the relationship.
            reflexivity (bool or None): the reflexivity of the relationship.
            complementary (string or None): if any, the obo_name of the
                complementary relationship.
            direction (string or None): if any, the direction of the
                relationship (can be 'topdown', 'bottomup', 'horizontal').
                A relationship with a direction set as 'topdown' will be
                counted as _childhooding_ when using the Term.children
                property.
        """
        if obo_name not in self._instances.keys():
            self.obo_name = obo_name
            self.symmetry = symmetry
            self.transitivity = reflexivity
            self.reflexivity = reflexivity
            self.complementary = complementary or ''
            self.prefix = prefix or ''
            self.direction = direction or ''
            self.comment = comment or ''
            self.aliases = aliases or []
            self._instances[obo_name] = self
            for alias in self.aliases:
                self._instances[alias] = self

    def complement(self):
        """Return the complementary relationship of self.

        Raises:
            ValueError: if the relationship has a complementary
                        which was not defined.

        Returns:
            complementary (Relationship): the complementary relationship.

        Example:

            >>> from pronto.relationship import Relationship
            >>> print(Relationship('has_part').complement())
            Relationship(part_of)
            >>> print(Relationship('has_units').complement())
            None

        """

        if self.complementary:

            if self.complementary in self._instances.keys():
                return self._instances[self.complementary]
            else:
                raise ValueError('{} has a complementary but it was not defined !')

        else:
            return None

    def __repr__(self):
        return "Relationship({})".format(self.obo_name)

    def __new__(cls, obo_name, *args, **kwargs):
        """Overloaded py:method::__new__ method that _memoizes_ the objects.

        This allows to do the following (which is frecking cool):

            >>> Relationship('has_part').direction
            'topdown'

        The Python syntax is overloaded, and what looks like a object
        initialization in fact retrieves an existing object with all its
        properties already set ! The Relationship class behaves like a
        factory of its own objects !

        Todo:
            * Add a warning for unknown relationship (the goal being to
              instantiate every known ontology relationship and even
              allow instatiation of file-defined relationships).

        """
        if obo_name in cls._instances:
            return cls._instances[obo_name]
        else:
            return super(Relationship, cls).__new__(cls)




Relationship('is_a', symmetry=False, transitivity=True,
                    reflexivity=True, complementary='can_be',
                    direction='bottomup')

Relationship('can_be', symmetry=False, transitivity=True,
                    reflexivity=True, complementary='is_a',
                    direction='topdown')

Relationship('has_part', symmetry=False, transitivity=True,
                        reflexivity=True, complementary='part_of',
                        direction='topdown')

Relationship('part_of', symmetry=False, transitivity=True,
                        reflexivity=True, complementary='has_part',
                        direction='bottomup', aliases=['is_part'])

Relationship('has_units', symmetry=False, transitivity=False,
                         reflexivity=None)





# coding: utf-8
"""
pronto.utils
============

This module contains some functions that are used in different parts
of the pronto library, as well as the definition of ProntoWarning.

Todo:
    * Maybe add a ProntoError class ?
"""


#def memoize(obj):
#    cache = obj.cache = {}
#
#    @functools.wraps(obj)
#    def memoizer(*args, **kwargs):
#        if args not in cache:
#            cache[args] = obj(*args, **kwargs)
#        return cache[args]
#
#    return memoizer



import functools
import errno
import os
import signal


class TimeoutError(IOError):
    pass


def timeout(seconds=60, error_message=os.strerror(errno.ETIME)):

    def decorator(func):

        def _handle_timeout(signum, frame):
            raise TimeoutError(error_message)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):

            try:
                signal.signal(signal.SIGALRM, _handle_timeout)
                signal.alarm(seconds)
            except ValueError:
                pass
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result
        return wrapper

    return decorator


class ProntoWarning(Warning):
    """A warning raised by pronto.

    Example:

        >>> from pronto import Ontology
        >>> import warnings
        >>> with warnings.catch_warnings(record=True) as w:
        ...    # the following ontology always has import issues (no URI)
        ...    ims = Ontology('https://raw.githubusercontent.com/beny/imzml/master/data/imagingMS.obo')
        >>> print(w[-1].category)
        <class 'pronto.utils.ProntoWarning'>

    """
    pass


def explicit_namespace(attr, nsmap):
    """Explicits the namespace in an attribute name.

    Example:

        >>> from pronto.utils import explicit_namespace
        >>> ns = {'owl':'http://www.w3.org/2002/07/owl#'}
        >>> explicit_namespace('owl:Class', ns)
        '{http://www.w3.org/2002/07/owl#}Class'

    """
    prefix, term = attr.split(':', 1)
    return '{'+nsmap[prefix]+'}'+term

def parse_comment(comment):
    """Parse an rdfs:comment to extract information.

    Owl contains comment which can contain additional metadata (specially
    when the Owl file was converted from Obo to Owl). This function parses
    the comment to try to extract those metadata.

    Arguments:
        - comment (str): if containing different sections (such as 'def:',
            'functional form' or 'altdef:'), the value of those sections will
            be returned in a dictionnary. If there are not sections, the
            comment is interpreted as a description

    Todo:
        * Add more parsing special cases

    """
    if comment is None:
        return {}

    commentlines = comment.split('\n')
    parsed = {}

    for (index, line) in enumerate(commentlines):

        line = line.strip()

        if line.startswith('Functional form:'):
            if not 'other' in parsed.keys():
                parsed['other'] = {}
            parsed['other']['functional form'] = "\n".join(commentlines[index:])
            break

        if line.startswith('def:'):
            parsed['desc'] = line.split('def:')[-1].strip()

        elif ': ' in line:
            ref, value = line.split(': ', 1)
            if not 'other' in parsed.keys():
                parsed['other'] = {}
            if not ref in parsed['other']:
                parsed['other'][ref.strip()] = []
            parsed['other'][ref.strip()].append(value)

        else:
            if not 'desc' in parsed.keys():
                parsed['desc'] = "\n".join(commentlines[index:])
                break

    if not 'desc' in parsed.keys() and 'other' in parsed.keys():
        if 'tempdef' in parsed['other'].keys():
            parsed['desc'] = parsed['other']['tempdef']
            del parsed['other']['tempdef']
        if 'altdef' in parsed['other'].keys():
            parsed['desc'] = parsed['other']['altdef']
            del parsed['other']['altdef']


    return parsed

def format_accession(accession, nsmap=None):
    """Formats an accession URI/string to the YY:XXXXXXX token format.

    Arguments:
        - accession (str): the URI to be formatted
        - nsmap (dict): namespaces that can be found at the beginning of the
            accession that we want to get rid of.

    Example:

        >>> from pronto.utils import format_accession
        >>> format_accession('UO_1000003')
        'UO:1000003'
        >>> ns = {'obo':'http://purl.obolibrary.org/obo/'}
        >>> format_accession('http://purl.obolibrary.org/obo/IAO_0000601', ns)
        'IAO:0000601'

    """

    if nsmap is not None:
        for v in nsmap.values():
            accession = accession.replace(v, '')

    if not accession.startswith('_'):
        accession = accession.replace('_', ':')

    return accession







"""
pronto.parser

This module contains all the parsers used to parse an ontology. The base class
Parser can be derived to provide additional parsers to ontology files.
"""

import pronto.term

from pronto.parser import Parser
from pronto.relationship import RSHIPS


class OboParser(Parser):
    """A parser for the Obo format.
    """

    def __init__(self):
        super(OboParser, self).__init__()
        self.extension = '.obo'
        self._rawterms = list()
        self._typedef = list()
        self._meta = dict()

    def hook(self, *args, **kwargs):
        """Returns True if the file is an Obo file (extension is .obo)"""
        if 'path' in kwargs:
            return kwargs['path'].endswith(self.extension)

    def read(self, stream):
        """Read the stream and extract information in a 'raw' format."""
        self._rawterms, self._typedef, self._meta = list(), list(), dict()
        IN = 'meta'
        for line in stream:
            line = line.strip().decode('utf-8') \
                   if hasattr(line, 'decode') \
                   else line.strip()
            IN = self._check_section(line) or IN
            if ': ' in line:
                self._parse_line_statement(line, IN)

    def makeTree(self, pool):
        """Create the proper ontology Tree from raw terms"""
        for t in pool.map(_classify, self._rawterms):
            self.terms.update(t)

    def metanalyze(self):
        """Analyze metadatas extracted from the beginning of the file."""

        for key,value in self._meta.items():
            if key != 'remark':
                self._parse_statement(key, value)
            else:
                for remark in value:
                    self._parse_remark(remark)

    def manage_imports(self):
        """Get metadatas concerning imports."""
        if 'import' in self.meta.keys():
            self.imports = self.meta['import']
        self.imports = list(set(self.imports))


    def _parse_remark(self, remark):
        """Parse a remark and add results to self.meta."""
        if ': ' in remark:
            key_remark, value_remark = remark.split(': ', 1)
            if not key_remark in self.meta:
                self.meta[key_remark] = []
            self.meta[key_remark].append(value_remark)
        else:
            if 'remark' not in self.meta:
                self.meta['remark'] = []
            self.meta['remark'].append(remark)

    def _parse_statement(self, key, value):
        """Parse a ``key: value`` statement in the ontology file."""
        if not isinstance(value, list):
            value = [value]
        if not key in self.meta.keys():
            self.meta[key] = []
        self.meta[key].extend(value)


    def _get_dict_to_update(self, IN):
        """Returns the right dictionnary to use"""
        if IN=='meta':
            return self._meta
        elif IN=='term':
            return self._rawterms[-1]
        elif IN=='typedef':
            return self._typedef[-1]

    def _check_section(self, line):
        """Check if current section must be changed."""
        if '[Term]' in line:
            self._rawterms.append({})
            return 'term'
        elif '[Typedef]' in line:
            self._typedef.append({})
            return 'typedef'

    def _parse_line_statement(self, line, IN):
        k, v = line.split(': ', 1)
        to_update = self._get_dict_to_update(IN)
        if k not in to_update.keys():
            to_update[k] = v
        else:
            if not isinstance(to_update[k], list):
                to_update[k] = [to_update[k]]
            to_update[k].append(v)



def _classify(term):

    relations = {}

    if 'relationship' in term:
        _process_relationship(term)

    for rship in RSHIPS:
        relations = _extract_relationship(term, rship, relations)

    desc = _extract_description(term)

    tid, name = _extract_name_and_id(term)

    if not tid:
        return {}

    return {tid: pronto.term.Term(tid, name, desc, relations, term)}


def _process_relationship(term):

    if isinstance(term['relationship'], list):
        for r in term['relationship']:
            name, value = r.split(' ', 1)

            if name not in term.keys():
                term[name] = []
            term[name].append(value)

    else:
        name, value = term['relationship'].split(' ', 1)

        if name not in term.keys():
            term[name] = []
        term[name].append(value)

    del term['relationship']

def _extract_relationship(term, rship, relations):
    if rship in term.keys():
        if isinstance(term[rship], list):
            relations[rship] = [ x.split(' !')[0] for x in term[rship] ]
        else:
            relations[rship] = [ term[rship].split(' !')[0]]
        del term[rship]
    return relations

def _extract_description(term):
    if 'def' in term.keys():
        desc = term['def']
        del term['def']
        return desc
    else:
        return ''

def _extract_name_and_id(term):
    if 'id' not in term.keys():
        return '', ''
    if 'name' in term.keys():
        tid, name = term['id'], term['name']
        del term['name']
        del term['id']
        return tid, name
    else:
        tid = term['id'][0]
        name = ''
        del term['id']
        return tid, name


OboParser()




import functools
import os
import lxml.etree as etree

from pronto.parser import Parser
import pronto.utils


class OwlXMLParser(Parser):
    """A parser for the owl xml format.
    """

    def __init__(self):
        super(OwlXMLParser, self).__init__()
        self._tree = None
        self._ns = {}
        self.extensions = ('.owl', '.xml', '.ont')

    def hook(self, *args, **kwargs):
        """Returns True if the file is an Owl file (extension is .owl)"""
        if 'path' in kwargs:
            return os.path.splitext(kwargs['path'])[1] in self.extensions

    def read(self, stream):
        """
        Parse the content of the stream
        """
        self._tree = etree.parse(stream)
        self._ns = self._tree.find('.').nsmap
        if None in self._ns.keys():
            self._ns['base'] = self._ns[None]
            del self._ns[None]

    def makeTree(self, pool):
        """
        Maps :function:_classify to each term of the file via a ThreadPool.

        Once all the raw terms are all classified, the :attrib:terms dictionnary
        gets updated.

        Arguments:
            pool (Pool): a pool of workers that is used to map the _classify
                function on the terms.
        """
        terms_elements = self._tree.findall('./owl:Class', self._ns)
        for t in pool.map(self._classify, terms_elements):
            self.terms.update(t)

    def _classify(self, term):
        """
        Map raw information extracted from each owl Class.

        The raw data (in an etree.Element object) is extracted to a proper
        dictionnary containing a Term referenced by its id, which is then
        used to update :attribute:terms

        Todo:
            * Split into smaller methods to lower code complexity.
        """

        nspaced = functools.partial(pronto.utils.explicit_namespace, nsmap=self._ns)
        accession = functools.partial(pronto.utils.format_accession, nsmap=self._ns)

        if not term.attrib:
           return {}

        tid = accession(term.get(nspaced('rdf:about')))

        term_dict = {'name':'', 'relations': {}, 'desc': ''}

        translator = [
            {'hook': lambda c: c.tag == nspaced('rdfs:label'),
             'callback': lambda c: c.text,
             'dest': 'name',
             'action': 'store'
            },
            {
             'hook': lambda c: c.tag == nspaced('rdfs:subClassOf') \
                               and nspaced('rdf:resource') in c.attrib.keys(),
             'callback': lambda c: accession(c.get(nspaced('rdf:resource')) or c.get(nspaced('rdf:about'))),
             'dest': 'relations',
             'action': 'list',
             'list_to': 'is_a'
            },
            {'hook': lambda c: c.tag == nspaced('rdfs:comment'),
             'callback': lambda c: pronto.utils.parse_comment(c.text),
             'action': 'update'
            }
        ]

        for child in term.iterchildren():

            for rule in translator:

                if rule['hook'](child):

                    if rule['action'] == 'store':
                        term_dict[rule['dest']] = rule['callback'](child)

                    elif rule['action'] == 'list':

                        if not term_dict[rule['dest']]:
                            term_dict[rule['dest']][rule['list_to']] = []

                        term_dict[rule['dest']][rule['list_to']].append(rule['callback'](child))


                    elif rule['action'] == 'update':
                        term_dict.update(rule['callback'](child))


                    break

        #if ':' in tid: #remove administrative classes
        return {tid: pronto.term.Term(tid, **term_dict)}
        #else:
        #    return {}

    def manage_imports(self):
        nspaced = functools.partial(pronto.utils.explicit_namespace, nsmap=self._ns)
        for imp in self._tree.iterfind('./owl:Ontology/owl:imports', self._ns):
            path = imp.attrib[nspaced('rdf:resource')]
            if path.endswith('.owl'):
                self.imports.append(path)

    def metanalyze(self):
        """
        Extract metadata from the headers of the owl file.

        Todo:
            * Implement that method !
        """
        pass


OwlXMLParser()



# coding: utf-8
"""
pronto.ontology
===============

This submodule contains the definition of the Ontology class.
"""


import json
import os
import warnings

import multiprocessing.dummy

try:
    import urllib.request as rq
    from urllib.error import URLError, HTTPError
except ImportError:
    import urllib2 as rq
    from urllib2 import URLError, HTTPError
    from pronto.utils import TimeoutError


import pronto.term
import pronto.parser
import pronto.utils


class Ontology(object):
    """The base class for an ontology.

    Ontologies inheriting from this class will be able to use the same API as
    providing they generated the expected structure in the :func:`_parse`
    method.

    Examples:
        Import an ontology from a remote location:

            >>> from pronto import Ontology
            >>> envo = Ontology("https://raw.githubusercontent.com/"
            ... "EnvironmentOntology/envo/master/src/envo/envo.obo")

        Merge two local ontologies and export the merge:

            >>> uo = Ontology("resources/uo.owl", False)
            >>> cl = Ontology("resources/cl.ont", False)

        Export an ontology with its dependencies embedded:

            >>> import os
            >>> cl = Ontology("resources/cl.ont")
            >>> with open('run/cl.obo', 'w') as f:
            ...     lines_count = f.write(cl.obo)


    Todo:
        * Add a __repr__ method
    """

    def __init__(self, path=None, imports=True, import_depth=-1):
        """
        """

        self.pool = multiprocessing.dummy.Pool()

        self.path = path
        self.meta = dict()
        self.terms = dict()
        self.imports = list()

        if path is not None:


            if path.startswith('http') or path.startswith('ftp'):
                handle = rq.urlopen(path)
            else:
                if not os.path.exists(path):
                    raise OSError('Ontology file {} could not be found'.format(path))
                else:
                    handle = open(path, 'r')


            self.parse(handle)

            self.adopt()

            if imports and import_depth:
                self.resolve_imports(import_depth)

            self.reference()

        self.pool.close()
        self.pool.join()
        del self.pool

    @property
    def json(self):
        """Returns the ontology serialized in json format.

        Example:
            >>> j = uo.json
            >>> all(term.id in j for term in uo)
            True

        Note:
            It is possible to save and load an ontology to and from
            json format, although it is cleaner to save and load
            an ontology in Obo format (as the json export doesn't store
            metadata, only terms).

        """
        return json.dumps(self.terms, indent=4, sort_keys=True,
                          default=lambda o: o.__deref__.__dict__)

    @property
    def obo(self):
        """Returns the ontology serialized in obo format.
        """
        obo = ""

        for accession in sorted(self.terms.keys()):
            obo += '\n'
            obo += self.terms[accession].obo

        return obo

    def reference(self):
        """Make relationships point to classes of ontology instead of ontology id"""
        for termkey,termval in self.terms.items():
            for relkey, relval in termval.relations.items():

                relvalref = [self.terms[x] if x in self.terms.keys()
                             else pronto.term.Term(x, '','') if not isinstance(x, pronto.term.Term)
                             else x for x in relval]
                self.terms[termkey].relations[relkey] = pronto.term.TermList(relvalref)

    def parse(self, stream):
        for parser in pronto.parser.Parser.instances.values():
            if parser.hook(stream=stream, path=self.path):
                try:
                    self.meta, self.terms, self.imports = parser.parse(stream, self.pool)
                    return
                except TimeoutError:
                    warnings.warn("Parsing of {} timed out".format(self.path),
                                   pronto.utils.ProntoWarning)



    def __getitem__(self, item):
        return self.terms[item]

    def __contains__(self, item):
        if isinstance(item, pronto.term.Term):
            return item.id in self.terms.keys()
        elif isinstance(item, str) or isinstance(item, unicode):
            return item in self.terms.keys()
        else:
            raise TypeError("'in <ontology>' requires string or Term as left operand, not {}".format(type(item)))

    def __iter__(self):
        terms_accessions = sorted(self.terms.keys())
        return (self.terms[i] for i in terms_accessions)

    def __len__(self):
        return len(self.terms)

    def adopt(self):
        """Make terms aware of their children via 'can_be' and 'has_part' relationships"""

        relationships = []

        for term in self:

            if 'is_a' in term.relations.keys():
                for parent in term.relations['is_a']:
                    relationships.append( (parent, 'can_be', term.id ) )

            if 'part_of' in term.relations.keys():
                for parent in term.relations['part_of']:
                    relationships.append( (parent, 'has_part', term.id ) )

            if 'is_part' in term.relations.keys():
                for parent in term.relations['is_part']:
                    relationships.append( (parent, 'part_of', term.id ) )

        for parent, rel, child in relationships:
            if isinstance(parent, pronto.term.Term):
                parent = parent.id
            if parent in self:
                if not rel in self[parent].relations.keys():

                    self[parent].relations[rel] = pronto.term.TermList()

                self[parent].relations[rel].append(child)

    def include(self, *terms):
        """Add new terms to the current ontology.

        Raises:
            TypeError: when the object(s) passed to the function is (are)
                not either a TermList or a Term.

        Note:
            This will also recursively include terms in the term's relations
            dictionnary, but it is considered bad practice to do so. If you
            want to create your own ontology, you should only add an ID (such
            as 'ONT:001') to your terms relations, and let the Ontology link
            terms with each other.

        :Example:

        Create a new ontology from scratch:

        >>> from pronto import Term
        >>> t1 = Term('ONT:001','my 1st term',
        ...           'this is my first term')
        >>> t2 = Term('ONT:002', 'my 2nd term',
        ...           'this is my second term',
        ...           {'part_of': ['ONT:001']})
        >>> ont = Ontology()
        >>> ont.include(t1, t2)

        >>> 'ONT:002' in ont
        True
        >>> ont['ONT:001'].children
        [<ONT:002: my 2nd term>]

        """
        ref_needed = False
        for term in terms:
            if isinstance(term, pronto.term.TermList):
                ref_needed = ref_needed or self._include_term_list(term)
            elif isinstance(term, pronto.term.Term):
                ref_needed = ref_needed or self._include_term(term)
            else:
                raise TypeError('include only accepts <Term> or <TermList> as arguments')

        self.adopt()
        self.reference()

    def resolve_imports(self, import_depth):
        """Imports required ontologies."""
        for i in set(self.imports):
            try:

                if os.path.exists(i) or i.startswith('http') or i.startswith('ftp'):
                    self.merge(Ontology(i, import_depth=import_depth-1))


                else: # try to look at neighbouring ontologies
                    self.merge(Ontology( os.path.join(os.path.dirname(self.path), i),
                                         import_depth=import_depth-1))

            except (IOError, OSError, URLError, HTTPError) as e:
                warnings.warn("{} occured when during import of "
                              "{}".format(type(e).__name__, i),
                              pronto.utils.ProntoWarning)

    def _include_term_list(self, termlist):
        """Add terms from a TermList to the ontology.
        """
        ref_needed = False
        for term in termlist:
            ref_needed = ref_needed or self._include_term(term)
        return reference_needed

    def _include_term(self, term):
        """Add a single term to the current ontology

        It is needed to dereference any term in the term's relationship
        and then to build the reference again to make sure the other
        terms referenced in the term's relations are the one contained
        in the ontology (to make sure changes to one term in the ontology
        will be applied to every other term related to that term).
        """
        ref_needed = False

        if term.relations:

            for k,v in term.relations.items():
                for i,t in enumerate(v):

                    if isinstance(t, pronto.term.Term):

                        if not t.id in self:
                            self._include_term(t)

                        term.relations[k][i] = t.id

                    ref_needed = True

        self.terms[term.id] = term
        return ref_needed

    def merge(self, other):
        """Merges another ontology into the current one.

        :Example:

        >>> from pronto import Ontology
        >>> nmr = Ontology('http://nmrml.org/cv/v1.0.rc1/nmrCV.owl', False)
        >>> po = Ontology('https://raw.githubusercontent.com/Planteome'
        ... '/plant-ontology/master/po.obo', False)
        >>> 'NMR:1000271' in nmr
        True
        >>> 'NMR:1000271' in po
        False
        >>> po.merge(nmr)
        >>> 'NMR:1000271' in po
        True

        """
        if isinstance(other, Ontology):
            self.terms.update(other.terms)
            self.reference()
        else:
            raise TypeError("'merge' requires an Ontology as argument, not {}".format(type(other)))




import pronto.utils

__all__ = ["Parser", "OboParser", "OwlXMLParser"]

class Parser(object):
    """A basic parser object.


    The following functions need to
    """

    instances = {}

    def __init__(self, timeout=None):
        self.terms = dict()
        self.meta = dict()
        self.imports = list()
        self.instances[type(self).__name__] = self

    @pronto.utils.timeout(500)
    def parse(self, stream, pool):
        """
        Parse the ontology file.

        :param stream: A stream of the ontology file.
        :type stream: io.StringIO
        """

        self.terms, self.meta, self.imports = dict(), dict(), list()

        self.read(stream)
        self.makeTree(pool)
        self.metanalyze()
        self.manage_imports()

        return self.meta, self.terms, self.imports

    def hook(self, path):
        """Defines when the parser is used

        For Obo and Owl, based on file extension (altough that may need to
        change depending on Owl format).
        """
        raise NotImplementedError

    def read(self, stream):
        """Reads and preprocesses the stream.
        """
        raise NotImplementedError

    def makeTree(self, pool):
        raise NotImplementedError

    def metanalyze(self):
        raise NotImplementedError

    def manage_imports(self):
        raise NotImplementedError




from pronto.parser.obo import OboParser
from pronto.parser.owl import OwlXMLParser



# coding: utf-8
"""
pronto.term
===========

This module defines the classes Term and TermList.
"""

#import functools

#from pronto import utils
from pronto.relationship import RSHIPS


class Term(object):
    """ An ontology term.
    """

    def __init__(self, tid, name, desc='', relations=None, other=None):
        self.id = tid
        self.name = name
        self.desc = desc
        self.relations = relations or {}
        self.other = other or {}

    def __repr__(self):
        return "<{}: {}>".format(self.id, self.name)

    @property
    #@functools.lru_cache(None)
    def parents(self):
        parents = TermList()
        for parental_rship in ('is_a', 'is_part', 'part_of'):
            if parental_rship in self.relations.keys():
                parents.extend(self.relations[parental_rship])
        return parents

    @property
    #@utils.memoize
    def children(self):
        children = TermList()
        for children_rship in ('has_part', 'can_be'):
            if children_rship in self.relations.keys():
                children.extend(self.relations[children_rship])
        return children

    @property
    def obo(self):

        obo =  '[Term]' + '\n'
        obo += 'id: {}'.format(self.id) + '\n'
        obo += 'name: {}'.format(self.name) + '\n'

        if self.desc:
            obo += 'def: {}'.format(self.desc) + '\n'

        # add more bits of information
        for k,v in self.other.items():
            if isinstance(v, list):
                for x in v:
                    obo += '{}: {}'.format(k,x) + '\n'
            else:
                obo += '{}: {}'.format(k,v) + '\n'

        # add relationships (only bottom up ones)
        for relation,companions in self.relations.items():
            if relation in RSHIPS:

                for companion in companions:

                    if relation != 'is_a':
                        obo += 'relationship: '
                    obo += '{}: '.format(relation)

                    if isinstance(companion, Term):
                        obo += '{} ! {}'.format(companion.id, companion.name) + '\n'
                    else:
                        obo += '{}'.format(companion)
                        obo += '\n'

        return obo

    @property
    def __deref__(self):
        """A dereferenced relations dictionary only contains other Terms id
        to avoid circular references when creating a json.
        """
        return Term(
            self.id,
            self.name,
            self.other,
            self.desc,
            {k:v.id for k,v in self.relations.items()}
         )

    def rchildren(self, level=-1, intermediate=True):
        """Create a recursive list of children.

        Note that the :param:`intermediate` can be used to include every
        child to the returned list, not only the most nested ones.

        """
        rchildren = []

        if level==0:
            return []

        if self.children:

            if intermediate or level==1:
                rchildren.extend(self.children)

            for child in self.children:
                rchildren.extend(child.rchildren(level=level-1,
                                                 intermediate=intermediate))

        return TermList(set(rchildren))

    def rparents(self, level=-1, intermediate=True):
        """Create a recursive list of parents.

        Note that the :param:`intermediate` can be used to include every
        parent to the returned list, not only the top ones.

        """
        rparents = []

        if level==0:
            return []

        if self.parents:

            if intermediate or level==1:
                rparents.extend(self.parents)

            for parent in self.parents:
                rparents.extend(parent.rparents(level=level-1,
                                                 intermediate=intermediate))

        return TermList(set(rparents))

class TermList(list):
    """A list of Terms.

    TermList behaves exactly like a list, except it contains shortcuts to
    generate lists of terms' attributes.

    Example:

        >>> from pronto import Ontology
        >>> nmr = Ontology('http://nmrml.org/cv/v1.0.rc1/nmrCV.owl')
        >>> type(nmr['NMR:1000031'].children)
        <class 'pronto.term.TermList'>

        >>> nmr['NMR:1000031'].children.id
        ['NMR:1000122', 'NMR:1000156', 'NMR:1000157', 'NMR:1000489']
        >>> nmr['NMR:1400014'].relations['is_a']
        [<NMR:1400011: cardinal part of NMR instrument>]


    .. note::
        It is also possible to call Term methods on a TermList to
        create a set of terms::

            >>> nmr['NMR:1000031'].rchildren(3, False).rparents(3, False).id
            ['NMR:1000031']

    """

    def __init__(self, *elements):
        list.__init__(self, *elements)
        self._check_content()

    def _check_content(self):
        for term in self:
            if not isinstance(term, Term):
                raise TypeError('TermList can only contain Terms.')

    def __getattr__(self, attr, *args, **kwargs):
        if attr in ['children', 'parents']:
            return TermList( [ y for x in self for y in getattr(x, attr)] )
        elif attr in ['rparents', 'rchildren']:
            #: we create a new method to allow the user
            #: to use, for instance, ``x.rchildren(3).rparents(2)``
            #: (this actually behaves as if you mapped the method
            #: on all terms of the TermList)
            def mapped(level=-1, intermediate=True):
                t = TermList(set([ y for x in self
                        for y in getattr(x, attr)(level, intermediate) ]))
                return t
            return mapped
        elif attr in ['id', 'name', 'desc', 'other']:
            return [getattr(x, attr) for x in self]
        else:
            getattr(list, attr)

    def __contains__(self, term):
        """
        Todo: write doc & test
        """
        return term in self.id or any(t for t in self if t==term)





#!/usr/bin/env python
# released under the GNU General Public License version 3.0 (GPLv3)

from setuptools import setup, find_packages
import pronto

## SETUPTOOLS VERSION
setup(
    name='pronto',
    version=pronto.__version__,

    packages=find_packages(),

    py_modules=['pronto'],

    author= pronto.__author__,
    author_email= pronto.__author_email__,

    description="Python frontend to ontologies - a library to parse, create, browse and export ontologies.",
    long_description=open('README.rst').read(),

    install_requires=open('requirements.txt').read().splitlines(),
    extras_require = { extra:open('requirements-{}.txt'.format(extra)).read().splitlines()
                        for extra in ['doc'] },

    include_package_data=True,

    url='http://github.com/althonos/pronto',

    classifiers=[
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
    "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Operating System :: OS Independent",
    ],

    license="GPLv3",

    keywords = ['Bio-Informatics', 'Ontology', 'OBO', 'Owl', 'convert', 'parse'],

)

